import os
from threading import Lock
import hashlib
import shutil
import subprocess
import time
import threading

def dictadd(_dict, key, val=1, ret_post=True):
    pre = _dict.get(key)
    post = (pre or 0) + val
    _dict[key] = post
    if ret_post:
        return post
    else:
        return pre

class GitJobDir(object):
    def __init__(s, basedir=None, maxdirs=4):
        s.basedir = basedir
        s.maxdirs = maxdirs
        s.dirs_left = maxdirs
        s.lock = Lock()
        s.use_counts = {}
        s.unused = {}
        s.deferred_clean_delay = 30

        try:
            os.mkdir(basedir)
        except FileExistsError:
            pass

    def dirkey(repo, commit, **kwargs):
        _str = "%s::%s::%s" % (repo, commit, hash(frozenset(kwargs.items())))
        print(_str, hash(frozenset(kwargs.items())))
        return hashlib.md5(_str.encode("utf-8")).hexdigest()

    def path(s, dirkey):
        return os.path.join(s.basedir, dirkey)

    def get(s, repo, commit, **kwargs):
        with s.lock:

            _dir = s.path(GitJobDir.dirkey(repo, commit, **kwargs))

            if dictadd(s.use_counts, _dir, ret_post=False) == None:
                if s.dirs_left == 0:
                    if not s.clean_unused():
                        s.use_counts.pop(_dir)
                        print("gitjobdir: could not get free jobdir slot")
                        return None
                try:
                    s.dirs_left -= 1
                    try:
                        s.lock.release()
                        s.checkout(repo, commit, **kwargs)
                        s.lock.acquire()
                    except Exception as e:
                        s.lock.acquire()
                        raise e
                except subprocess.CalledProcessError:
                    return None
            else:
                lock = s.unused.pop(_dir, None)
                if lock:
                    lock.release()

            return _dir

    def clean_unused(s):
        for _dir, lock in s.unused.items():
            if s.clean_dir(_dir):
                print("gitjobdir: randomly cleaned", _dir)
                lock.release()
                return True
        return False

    def release(s, _dir):
        with s.lock:
            use_count = s.use_counts.get(_dir)
            if use_count == 0 or use_count == None:
                print("GitJobDir: warning: release() on unused job dir!")
            else:
                if dictadd(s.use_counts, _dir, -1) == 0:
                    print("GitJobDir: last user of %s gone." % _dir)
                    s.clean_deferred(_dir)

    def clean_dir(s, _dir, delete_only=False):
        print("gitjobdir: cleaning directory", _dir)
        if not delete_only:
            s.unused.pop(_dir, None)
            s.use_counts.pop(_dir, None)
            s.dirs_left += 1
        shutil.rmtree(_dir)
        return True

    def clean_deferred(s, _dir):
        lock = Lock()
        lock.acquire()
        s.unused[_dir] = lock
        threading.Thread(target=GitJobDir.clean_deferred_handler, args=(s, _dir, lock)).start()

    def clean_deferred_handler(s, _dir, lock):
        print("clean_deferred_handler() waiting for", _dir)
        lock.acquire(timeout=s.deferred_clean_delay)
        with s.lock:
            if s.unused.get(_dir) is lock:
                print("clean_deferred_handler() triggered for", _dir)
                s.clean_dir(_dir)

    def checkout(s, repo, commit, **kwargs):
        target_path = s.path(GitJobDir.dirkey(repo, commit, **kwargs))
        subprocess.check_call(["git", "cache", "clone", repo, commit, target_path])

    def cleanup(s):
        with s.lock:
            for _dir, v in s.use_counts.items():
                try:
                    s.clean_dir(_dir, True)
                except FileNotFoundError:
                    pass

if __name__=="__main__":
    gjd = GitJobDir("/tmp/gitjobdir", maxdirs=1)

    _dira = gjd.get("http://github.com/RIOT-OS/RIOT", "c879154d144a349f890ab74ca8e0c70ded359de8")
    print("got", _dira)
    _dirb = gjd.get("http://github.com/RIOT-OS/RIOT", "96ef13dc9801595f4aec8763bd9c36278b5789e9")
    print("got", _dirb)
    print("releasing first")
    gjd.release(_dira)
    if not _dirb:
        print("trying second again", _dirb)
        _dirb = gjd.get("http://github.com/RIOT-OS/RIOT", "96ef13dc9801595f4aec8763bd9c36278b5789e9")
        print("got", _dirb)

    print("releasing 2nd")
    gjd.release(_dirb)

    _exclusive = gjd.get("http://github.com/RIOT-OS/RIOT", "96ef13dc9801595f4aec8763bd9c36278b5789e9", "TEST")
    gjd.release(_exclusive)
    time.sleep(1)

    _dirb = gjd.get("http://github.com/RIOT-OS/RIOT", "96ef13dc9801595f4aec8763bd9c36278b5789e9")
    print("got", _dirb)

    time.sleep(5)
    gjd.cleanup()



#!/usr/bin/env python3

import argparse
import os
import threading
import random
import time
import signal
import socket
import traceback
import multiprocessing

import redis
from redis.exceptions import ConnectionError, RedisError

import cmdserver
from dwq import Job, Disque

from gitjobdir import GitJobDir

class GracefulExit(Exception):
    pass

def sigterm_handler(signal, stack_frame):
    raise GracefulExit()

def parse_args():
    parser = argparse.ArgumentParser(prog='dwqw', description='dwq: disque-based work queue (worker)')

    parser.add_argument('--version', action='version', version='%(prog)s 0.1')

    parser.add_argument('-q', '--queues', type=str,
            help='queues to wait for jobs (default: \"default\")', nargs='*', default=["default"])
    parser.add_argument('-j', "--jobs",
            help='number of workers to start', type=int, default=multiprocessing.cpu_count())
    parser.add_argument('-n', '--name', type=str,
            help='name of this worker (default: hostname)', default=socket.gethostname())
    parser.add_argument('-v', "--verbose", help='enable status output', action="store_true" )

    return parser.parse_args()

shutdown = False

def worker(n, cmd_server_pool, gitjobdir, args, working_set):
    global shutdown
    print("worker %2i: started" % n)

    while not shutdown:
        try:
            if not Disque.connected():
                time.sleep(1)
                continue
            while not shutdown:
                jobs = Job.get(args.queues)
                for job in jobs:
                    if shutdown:
                        job.nack()
                        continue
                    working_set.add(job.job_id)
                    before = time.time()
                    print("worker %2i: got job %s from queue %s" % (n, job.job_id, job.queue_name))

                    try:
                        repo = job.body["repo"]
                        commit = job.body["commit"]
                        command = job.body["command"]
                    except KeyError:
                        print("worker %2i: invalid job json body" % n)
                        job.done({ "status" : "error", "output" : "worker.py: invalid job description" })
                        continue

                    exclusive = None
                    try:
                        options = job.body.get("options") or {}
                        if options.get("jobdir") or "" == "exclusive":
                            exclusive = str(random.random())
                    except KeyError:
                        pass

                    _env = os.environ.copy()
                    _env.update({ "DWQ_REPO" : repo, "DWQ_COMMIT" : commit, "DWQ_QUEUE" : job.queue_name })
                    try:
                        _env.update(job.body["env"])
                    except KeyError:
                        pass

                    workdir = gitjobdir.get(repo, commit, exclusive=exclusive or str(n))
                    if not workdir:
                        if job.nacks < (options.get("max_retries") or 2):
                            job.nack()
                            print("worker %2i: cannot get job dir, requeueing job" % n)
                        else:
                            job.done({ "status" : "error", "output" : "dwqw: error getting jobdir\n",
                                        "worker" : args.name, "runtime" : 0, "body" : job.body })
                            print("worker %2i: cannot get job dir, erroring job" % n)
                        working_set.discard(job.job_id)
                        continue

                    handle = cmd_server_pool.runcmd(command, cwd=workdir, shell=True, env=_env)
                    output, result = handle.wait(timeout=300)
                    if handle.timeout:
                        result = "timeout"

                    if (result not in { 0, "0", "pass" }) and job.nacks < (options.get("max_retries") or 2):
                        print("worker %2i: result:" % n, result, "nacks:", job.nacks, "re-queueing.")
                        job.nack()
                    else:
                        runtime = time.time() - before
                        job.done({ "status" : result, "output" : output, "worker" : args.name, "runtime" : runtime, "body" : job.body })
                        print("worker %2i: result:" % n, result, "runtime: %.1fs" % runtime)
                        working_set.discard(job.job_id)
                    gitjobdir.release(workdir)

        except Exception as e:
            print("worker %2i: uncaught exception" % n)
            traceback.print_exc()
            time.sleep(10)
            print("worker %2i: restarting worker" % n)


class SyncSet(object):
    def __init__(s):
        s.set = set()
        s.lock = threading.Lock()

    def add(s, obj):
        with s.lock:
            s.set.add(obj)

    def discard(s, obj):
        with s.lock:
            s.set.discard(obj)

    def empty(s):
        with s.lock:
            oldset = s.set
            s.set = set()
            return oldset

def main():
    global shutdown

    args = parse_args()

    signal.signal(signal.SIGTERM, sigterm_handler)

    cmd_server_pool = cmdserver.CmdServerPool(args.jobs)

    _dir = "/tmp/dwq.%s" % str(random.random())
    gitjobdir = GitJobDir(_dir, args.jobs)

    servers = ["localhost:7711"]
    try:
        Disque.connect(servers)
        print("dwqw: connected.")
    except:
        pass

    working_set = SyncSet()

    for n in range(1, args.jobs + 1):
        threading.Thread(target=worker, args=(n, cmd_server_pool, gitjobdir, args, working_set), daemon=True).start()

    try:
        while True:
            if not Disque.connected():
                try:
                    print("dwqw: connecting...")
                    Disque.connect(servers)
                    print("dwqw: connected.")
                except RedisError:
                    pass
            time.sleep(1)
    except (KeyboardInterrupt, GracefulExit):
        print("dwqw: signal caught, shutting down")
        shutdown = True
        cmd_server_pool.destroy()
        print("dwqw: nack'ing jobs")
        jobs = working_set.empty()
        d = Disque.get()
        d.nack_job(*jobs)
        print("dwqw: cleaning up job directories")
        gitjobdir.cleanup()




# python's subprocess creation speed depends on the main processes memory consumption
# (due to mmap'ing magic). So with a million targets, pyjam's overhead for calling "true"
# goes from 3ms to 20ms. Using this cmdserver class, the speed stays constant at 3.7ms.
# (measured on my thinkpad x220 core i5).
#
# The class also provides an easy API to run commands in a subprocess, gathering it's output,
# and provides a way of killing it.
# Part of this will be obsolete as soon as everyone can use Python 3.5 subprocess.run().
#

from multiprocessing import Process, Queue
from queue import Empty
from subprocess import Popen, PIPE, STDOUT
import signal
from collections import deque
import os

class CmdTimeout(Exception):
    pass

class CmdHandle(object):
    def __init__(s, queue, pid, pool, server):
        s.queue = queue
        s.pid = pid
        s.pool = pool
        s.server = server
        s.timeout = False

    def wait(s, timeout=None):
        try:
            res = s.queue.get(timeout=timeout)
        except Empty:
            s.timeout = True
            return s.kill()

        s.pool.append(s.server)
        return res

    def kill(s, signal=signal.SIGKILL):
        try:
            os.kill(-s.pid, signal)
        except:
            pass
        return s.wait()

    def killpg(s, signal=signal.SIGKILL):
        try:
            os.killpg(os.getpgid(s.pid), signal)
        except:
            pass
        return s.wait()

class CmdServer(object):
    def __init__(s, pool):
        s.inQueue = Queue()
        s.outQueue = Queue()
        s.cmdHostProcess = Process(target=CmdServer.cmdloop, args=(s, s.inQueue, s.outQueue, pool), daemon=True)
        s.cmdHostProcess.start()

    def cmdloop(s, inQueue, outQueue, pool):
        while True:
            args, kwargs = inQueue.get()
            kwargs["stderr"] = STDOUT
            kwargs["stdout"] = PIPE
            process = Popen(*args, **kwargs)
            outQueue.put(process.pid)
            output = u""
            with process:
                output += process.stdout.read().decode("utf-8", "replace")
            outQueue.put((output, process.returncode))

    def runcmd(s, *args, **kwargs):
        s.inQueue.put((args, kwargs))
        pid = s.outQueue.get()
        return s.outQueue, pid

    def killCmdHostProcess(s):
        s.cmdHostProcess.terminate()

class CmdServerPool(object):
    def __init__(s, n):
        s.pool = deque()
        for i in range(0, n):
            s.pool.append(CmdServer(s.pool))

    def destroy(s):
        while s.pool:
            s.pool.pop().killCmdHostProcess()

    def runcmd(s, *args, **kwargs):
        server = s.pool.pop()
        queue, pid = server.runcmd(*args, **kwargs)
        return CmdHandle(queue, pid, s.pool, server)
